/******************************************************************************
 * Kamil Raś                                                                  *
 * Grupa H1ISI                                                                *
 *                                                                            *
 * [ZPR] - Praca domowa, zadanie nr 8                                         *
 * Wzorzec projektowy: Adapter Obiektów (ang. object adapter, wrapper)        *
 ******************************************************************************/
/******************************************************************************
 * Wzorzec projektowy adapter polega na dostosowaniu interfejsu klasy do      *
 * interfejsu, którego oczekuje użytkownik. Adapter umożliwia współpracę      *
 * klas, która bez jego zastosowania nie byłaby możliwa ze względu na ich     *
 * niezgodne interfejsy.                                                      *
 * Wyróżnia się dwa rodzaje adapterów: adapter obiektów i adapter klas.       *
 * Zalety adaptera obiektów:                                                  *
 *   -> adapter obiektów jest bardziej elastyczny od adaptera klas - może     *
 *      zawierać nie tylko obiekty agregowane, ale tażke obiekty ich klas     *
 *      pochodnych;                                                           *
 *   -> może zaimplementować brakującą funkcjonalność agregowanego obiektu    *
 *      (to także może zrobić adapter klas);                                  *
 *   -> może agregować dwa lub więcej obiektów, żeby zaimplementować docelowy *
 *      interfejs;                                                            *
 * Wady adaptera obiektów:                                                    *
 *   -> wolniejszy w stosunku do adaptera klas - wprowadza dodatkowy poziom   *
 *      pośredniości;                                                         *
 *   -> trudniej nadpisać zachowanie klasy agregowanej (w stosunku do         *
 *      adaptera klas);                                                       *
 *                                                                            *
 * Korzystanie z adapterów może czasem powodować spore utrudnienia. Np. jeśli *
 * w systemie mamy zarówno nowe i stare moduły (stare wymagają posługiwania   *
 * się poprzednim, starym interfejsem) to korzystanie z adapterów może być    *
 * kłopotliwe - w jednej części aplikacji należy korzystać z odpowiedniego    *
 * adaptera, a w innej z bezpośrednich odwołań do interfejsu klasy            *
 * agregowanej. Należy wtedy utworzyć adapter zapewniający dwukierunkową      *
 * przezroczystość - taki, który będzie implementował oba przedmiotowe        *
 * iterfejsy, dzięki czemu nowy adapter będzie mógł odgrywać rolę zarówno     * 
 * starego, jak i nowego interfejsu.                                          *
 ******************************************************************************/

#include <iostream>
#include <cmath>

// Intefejs używany przez użytkownika
class IShape
{
public:
    virtual ~IShape() { }
    virtual void display() const = 0;
    virtual double getArea() const = 0;
};

/******************************************************************************/
// Przykładowa klasa niedostosowana do interfejsu używanego przez użytkownika
// posiadająca całą wymaganą przez użytkownika funkcjonalność
class LegacyRectangle
{
public:
    LegacyRectangle (double x1, double y1, double x2, double y2) :
        x1_(x1), y1_(y1), x2_(x2), y2_(y2) { }
    virtual ~LegacyRectangle() { }
    
    virtual void displayMe() const
    {
        std::cout << "Rectangle: (" << x1_ << ", " << y1_ << ") (" << x2_ 
            << ", " << y2_ << ")" << std::endl;
    }
    virtual double myArea() const
    {
        return (x2_ - x1_) * (y1_ - y2_);
    }

protected:
    // Współrzędne lewego górnego oraz prawego dolnego rogu
    double x1_;
    double y1_;
    double x2_;
    double y2_;
};

class LegacySquare : public LegacyRectangle
{
public:
    LegacySquare (double x1, double y1, double x2, double y2) :
        LegacyRectangle(x1, y1, x2, y2) { }
    ~LegacySquare() { }
    
    virtual void displayMe() const
    {
        std::cout << "Square: (" << x1_ << ", " << y1_ << ") (" << x2_ 
            << ", " << y2_ << ")" << std::endl;
    }
    virtual double myArea() const
    {
        return 0.5 * (pow(x2_ - x1_, 2) + pow(y1_ - y2_, 2));
    }
};

// Adapter dla obiektów LegacyRectangle. Deleguje wywoływanie funcji 
// z interfejsu IShape do obiektu agregowanego. Dzięki zastosowaniu adaptera
// obiektów może agregować także obiekty klas dziedziczących po LegacyRectangle,
// np. LegacySquare (adapter klas tego nie potrafi).
class Rectangle : public IShape
{
public:
    Rectangle(LegacyRectangle * rectangle) : rectangle_(rectangle) { }
    ~Rectangle()
    {
        delete rectangle_;
    }

    virtual void display() const
    {
        rectangle_->displayMe();
    }
    virtual double getArea() const
    {
        return rectangle_->myArea();
    }

private:
    LegacyRectangle * rectangle_;
};

/******************************************************************************/
// Klasa nie implementująca interfejsu używanego przez użytkownika, lecz
// posiadająca tylko część wymaganej funkcjonalności
class LegacyCircle
{
public:
    LegacyCircle(double x1, double y1, double radius) : x1_(x1), y1_(y1), 
        radius_(radius) { }
    ~LegacyCircle() { }

    void displayCircle() const
    {
        std::cout << "Circle: (" << x1_ << ", " << y1_ << "), R=" << radius_ 
            << std::endl;
    }
    double getRadius() const
    {
        return radius_;
    }
    static const double GetPI()
    {
        return PI;
    }
private:
    static const double PI;
    double x1_;
    double y1_;
    double radius_;
};
const double LegacyCircle::PI = 3.14;

// Adapter obietków LegacyCircle. Mimo, że LegacyCircle nie posiada całej
// funkcjonalności wymaganej przez IShape, adapter może tą "lukę" wypełnić
// poprzez zaimplementowanie brakujących funkcji.
class Circle : public IShape
{
public:
    Circle(LegacyCircle * circle) : circle_(circle) { }
    ~Circle()
    {
        delete circle_;
    }
    
    // Adapter deleguje wywołanie metody do obiektu agregowanego
    virtual void display() const
    {
        circle_->displayCircle();
    }
    // Jeśli obiekt agregowany nie posiada całej funkcjonalności, to w adapterze
    // możemy tą brakującą funkcjonalność zaimplementować
    virtual double getArea() const
    {
        return 2 * circle_->getRadius() * circle_->GetPI();
    }

private:
    LegacyCircle * circle_;
};

/******************************************************************************/
// Interfejs używany przez użytkownika do rysowania figur
class IDrawing
{
public:
    virtual ~IDrawing() { }
    virtual void drawBorder() = 0;
    virtual void drawFilling() = 0;
};

// Klasa posiadająca część funkcjonalności wymaganej przez użytkownika 
// (przez interfejs IDrawing)
class Square
{
public:
    Square(double x1, double y1, double x2, double y2) 
        : x1_(x1), y1_(y1), x2_(x2), y2_(y2) { }
    ~Square() { }

    void draw() const
    {
        std::cout << "Drawing empty square at (" << x1_ << ", " << y1_ 
            << "), (" << x2_ << ", " << y2_ << ")" << std::endl;
    }
    double getX1() const
    {
        return x1_;
    }
    double getY1() const
    {
        return y1_;
    }
    double getX2() const 
    {
        return x2_;
    }
    double getY2() const
    {
        return y2_;
    }
private:
    double x1_;
    double y1_;
    double x2_;
    double y2_;
};

// Klasa posiadająca część funkcjonalności wymaganej przez użytkownika 
// (przez interfejs IDrawing)
class SolidFilling
{
public:
    SolidFilling(double x1, double y1, double x2, double y2) 
        : x1_(x1), y1_(y1), x2_(x2), y2_(y2) { }
    ~SolidFilling() { }

    void drawSolidFilling()
    {
        std::cout << "Drawing solid filing at (" << x1_ << ", " << y1_
            << "), (" << x2_ << ", " << y2_ << ")" << std::endl;
    }

private:
    double x1_;
    double y1_;
    double x2_;
    double y2_;
};

// Adapter implementujący całość funkcjonalości wymaganej przez użytkownika
// (interfejs IDrawing) dzięki agregacji obiektów Square i SolidFilling, które 
// razem posiadają wymaganą funkcjonalność.
class DrawingSquare : public IDrawing
{
public:
    DrawingSquare(Square * square) : square_(square)
    {
        solidFiling_ = new SolidFilling(square_->getX1(), square_->getY1(), 
            square_->getX2(), square_->getY2());
    }
    ~DrawingSquare()
    {
        delete square_;
        delete solidFiling_;
    }

    virtual void drawBorder()
    {
        square_->draw();
    }
    virtual void drawFilling()
    {
        solidFiling_->drawSolidFilling();
    }
private:
    Square * square_;
    SolidFilling * solidFiling_;
};

/******************************************************************************/
int main()
{
    const int MAX = 3;
    IShape * shapes[MAX];
    shapes[0] = new Rectangle(new LegacyRectangle(1.0, 5.0, 2.0, 1.0));
    shapes[1] = new Rectangle(new LegacySquare(3.0, 5.0, 7.0, 1.0));
    shapes[2] = new Circle(new LegacyCircle(10.0, 10.0, 5.0));
    for (int i = 0; i < MAX; ++i)
    {
        shapes[i]->display();
        shapes[i]->getArea();
    }

    IDrawing * drawing = new DrawingSquare(new Square(20.5, 20.5, 30.5, 10.5));
    drawing->drawBorder();
    drawing->drawFilling();

    for (int i = 0; i < MAX; ++i)
        delete shapes[i];
    delete drawing;

    system("PAUSE");
    return 0;
}
